"""Prime Component"""

from ..objects.component import Activation, Component
from ..systems.execution import docker

class Prime(Component):
    """
    Prime Class

    This component provides the ability to prime a skelebot build by constructing the Dockerfile,
    the .dockerignore file, and building the docker image for the project
    """

    activation = Activation.PROJECT
    commands = ["prime"]

    def addParsers(self, subparsers):
        """
        SkeleParser Hook

        Adds a parser for the prime command that primes the docker image for deployment or
        subsequent job execution
        """

        helpMessage = "Generate Dockerfile and .dockerignore and build the docker image"
        primeParser = subparsers.add_parser("prime", help=helpMessage)
        primeParser.add_argument("-o", "--output", help="Output file name for the image file")
        return subparsers

    def execute(self, config, args, host=None):
        """
        Execution Hook

        When the prime command is provided the Dockerfile, dockerignore file, and docker image are
        built so as to be ready for deployment
        """

        status = docker.build(config, host=host, verbose=args.verbose_global)
        if (status == 0) & (args.output is not None):
            status = docker.save(config, args.output, host=host, verbose=args.verbose_global)

        if (status > 0):
            raise Exception("Priming Failed")



"""Scaffolder System"""

import os
import re
from subprocess import call

import yaml as pyyaml

from ...objects.config import Config
from ...components.componentFactory import ComponentFactory
from ...systems.generators import dockerfile, dockerignore, readme, yaml
from ...common import TEMPLATES, TEMPLATE_PATH
from .prompt import promptUser

class Scaffolder:

    existing = None
    variables = None

    def __format_variables(self, text):
        for key, value in self.variables.items():
            key_str = f"{{{key}}}"
            if key_str in text:
                text = text.replace(key_str, value)

        return text

    def __format_config(self, config):
        return eval(self.__format_variables(str(config)))

    def __load_git_template(self, url):
        # Construct template_name (folder name) and template path
        template_name = re.sub("[:/\\.@-]+", "_", url)
        template_path = TEMPLATE_PATH.format(name=template_name)
        template_path = os.path.join(os.path.dirname(__file__), template_path)

        # Clone or pull the template from the Git URL into the template folder
        if (os.path.exists(template_path)):
            # If the template has been cloned before, just pull --rebase
            status = call(f"cd {template_path} && git pull --rebase", shell=True)
        else:
            # If the template has NOT been cloned before, clone it
            status = call(f"git clone {url} {template_path}", shell=True)

        return self.__load_template(template_name)

    def __load_template(self, name):
        template = None

        path = TEMPLATE_PATH.format(name=name)
        path = os.path.join(os.path.dirname(__file__), path)
        with open(f"{path}/template.yaml", "r", encoding="utf-8") as yaml_file:
            yaml_text = self.__format_variables(yaml_file.read())
            template = pyyaml.load(yaml_text, Loader=pyyaml.FullLoader)

        # Update file paths in the template
        for file_dict in template.get("files", []):
            file_path = file_dict["template"]
            file_dict["template"] = f"{path}/{file_path}"

        return template

    def __init__(self, existing=True):
        self.existing = existing
        self.variables = {}

    def scaffold(self):
        """Scaffold a new Dashboard Project"""

        # Prompt for basic project information
        print("Scaffolding Skelebot Project")
        print("--:-" * 5, "-:--" * 5)
        name = promptUser("Enter a PROJECT NAME")
        description = promptUser("Enter a PROJECT DESCRIPTION")
        maintainer = promptUser("Enter a MAINTAINER NAME")
        contact = promptUser("Enter a CONTACT EMAIL")

        # Configure Template Variables
        self.variables["name"] = name
        self.variables["name_simple"] = name.lower().replace(" ", "_").replace("-", "_")
        self.variables["description"] = description
        self.variables["maintainer"] = maintainer
        self.variables["contact"] = contact

        # Load the Template Config
        options = list(TEMPLATES.keys())
        template = promptUser("Select a TEMPLATE", options=options)
        template_name = TEMPLATES[template.capitalize()]
        if (template_name == "git"):
            url = promptUser("Enter Git Repo URL")
            template = self.__load_git_template(url)
        else:
            template = self.__load_template(template_name)

        # Prompt user based on the template
        for prompt in template.get("prompts", []):
            self.variables[prompt["var"]] = promptUser(prompt["message"])

        # Iterate over components for additional prompts and add any components that are scaffolded
        components = []
        componentFactory = ComponentFactory()
        for component in componentFactory.buildComponents():
            component = component.scaffold()
            if (component is not None):
                components.append(component)

        # Confirm user input - allow them to back out before generating files
        print("--:-" * 5, "-:--" * 5)
        print("Setting up the", name, "Skelebot project in the current directory")
        print("(", os.getcwd(), ")")
        if (not promptUser("Confirm Skelebot Setup", boolean=True)):
            raise Exception("Aborting Scaffolding Process")

        print("--:-" * 5, "-:--" * 5)
        if (not self.existing):
            # Setting up the folder structure for the project
            print("Rigging up the skele-bones...")
            for directory in template.get("dirs", []):
                os.makedirs(directory, exist_ok=True)

            # Setting up the file templates for the project
            print("Attaching fiber optic ligaments...")
            for file_dict in template.get("files", []):
                with open(file_dict["template"], "r", encoding="utf-8") as tmp_file:
                    with open(file_dict["name"], "w", encoding="utf-8") as dst_file:
                        dst_file.write(self.__format_variables(tmp_file.read()))

        print("Soldering the micro-computer to the skele-skull...")
        # Build the config object based on the user inputs
        cfg = self.__format_config(template["config"])
        cfg["name"] = name
        if "components" not in cfg:
            cfg["components"] = {}
        for component in components:
            component_dict = component.toDict()
            if component_dict != {}:
                cfg["components"][component.__class__.__name__.lower()] = component_dict

        config = Config.load(cfg)
        config.description = description
        config.maintainer = maintainer
        config.contact = contact
        config.version = "0.1.0"
        yaml.saveConfig(config)

        if (not self.existing):
            # Creating the files for the project
            print("Initializing default systems...")
            dockerfile.buildDockerfile(config)
            dockerignore.buildDockerignore(config)
            readme.buildREADME(config)

        print("Your Skelebot project is ready to go!")



import os
import unittest
from unittest import mock

import skelebot as sb

class TestDockerignore(unittest.TestCase):
    path = ""
    config = None

    # Get the path to the current working directory before we mock the function to do so
    def setUp(self):
        self.path = os.getcwd()

    @mock.patch('os.path.expanduser')
    @mock.patch('os.getcwd')
    def test_buildDockerignore(self, mock_getcwd, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        filePath = "{folder}/.dockerignore".format(folder=folderPath)
        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        self.config = sb.systems.generators.yaml.loadConfig()

        expected = """
# This dockerignore was generated by Skelebot
# Editing this file manually is not advised as all changes will be overwritten by Skelebot

**/*.zip
**/*.pkl
**/*.csv
**/*.model
**/*.pyc
"""

        sb.systems.generators.dockerignore.buildDockerignore(self.config)

        data = None
        with open(filePath, "r", encoding="utf-8") as file:
            data = file.read()
        self.assertTrue(data is not None)
        self.assertEqual(data, expected)

if __name__ == '__main__':
    unittest.main()



import os
import unittest
from unittest import mock
from argparse import Namespace

from colorama import Fore, Style

import skelebot as sb

class LimitMemory(sb.objects.component.Component):
    activation = sb.objects.component.Activation.CONFIG

    def __init__(self, memory):
        self.memory = memory

    def addDockerRunParams(self):
        return "--memory {memory}GB".format(memory=self.memory)

class TestDocker(unittest.TestCase):
    path = ""

    # Get the path to the current working directory before we mock the function to do so
    def setUp(self):
        self.path = os.getcwd()

    @mock.patch('skelebot.systems.execution.docker.call')
    def test_login(self, mock_call):
        mock_call.return_value = 0
        mock_call.configure_mock(shell=True)

        sb.systems.execution.docker.login()
        mock_call.assert_called_once_with("docker login ", shell=True)

    @mock.patch('skelebot.systems.execution.docker.call')
    def test_login_host(self, mock_call):
        mock_call.return_value = 0

        sb.systems.execution.docker.login("docker.io")
        mock_call.assert_called_once_with("docker login docker.io", shell=True)

    @mock.patch('skelebot.systems.execution.docker.call')
    def test_login_error(self, mock_call):
        mock_call.return_value = 1

        with self.assertRaisesRegex(Exception, "Docker Login Failed"):
            sb.systems.execution.docker.login()

    @mock.patch('skelebot.systems.execution.docker.call')
    def test_login_aws_v2(self, mock_call):
        mock_call.return_value = 0

        sb.systems.execution.docker.loginAWS("123.dkr.ecr.us-east-1.amazonaws.com", "us-east-1", "dev")
        mock_call.assert_called_once_with("aws ecr get-login-password --region us-east-1 --profile dev | docker login --username AWS --password-stdin 123.dkr.ecr.us-east-1.amazonaws.com", shell=True)

    @mock.patch('skelebot.systems.execution.docker.call')
    def test_login_aws_v2_host(self, mock_call):
        mock_call.return_value = 0

        sb.systems.execution.docker.loginAWS("123.dkr.ecr.us-east-1.amazonaws.com", "us-east-1", "dev", "host1")
        mock_call.assert_called_once_with("aws ecr get-login-password --region us-east-1 --profile dev | docker -H host1 login --username AWS --password-stdin 123.dkr.ecr.us-east-1.amazonaws.com", shell=True)

    @mock.patch('skelebot.systems.execution.docker.call')
    def test_login_aws_v1(self, mock_call):
        mock_call.return_value = 1

        with self.assertRaisesRegex(Exception, "Docker Login V1 Failed"):
            sb.systems.execution.docker.loginAWS(None, "us-east-1", "dev")

        mock_call.return_value = 0

        sb.systems.execution.docker.loginAWS("123.dkr.ecr.us-east-1.amazonaws.com", "us-east-1", "dev")
        mock_call.assert_any_call("$(aws ecr get-login --no-include-email --region us-east-1 --profile dev)", shell=True)

    @mock.patch('skelebot.systems.execution.docker.call')
    def test_login_aws_v1_host(self, mock_call):
        mock_call.return_value = 1

        with self.assertRaisesRegex(Exception, "Docker Login V1 Failed"):
            sb.systems.execution.docker.loginAWS(None, "us-east-1", "dev")

        with self.assertRaisesRegex(Exception, "Remote hosts are not supported"):
            sb.systems.execution.docker.loginAWS(
                "123.dkr.ecr.us-east-1.amazonaws.com", "us-east-1", "dev", "host1"
            )

    @mock.patch('skelebot.systems.execution.docker.call')
    def test_login_aws_error(self, mock_call):
        mock_call.return_value = 1

        with self.assertRaisesRegex(Exception, r"Docker Login .* Failed"):
            sb.systems.execution.docker.loginAWS(None, "us-east-1", "dev")

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_push(self, mock_getcwd, mock_call, mock_expanduser):
        host = "docker.io"
        port = 8888
        user = "skelebot"
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        sb.systems.execution.docker.push(config, host, port, user)

        mock_call.assert_any_call("docker tag test docker.io:8888/skelebot/test:6.6.6", shell=True)
        mock_call.assert_any_call("docker tag test docker.io:8888/skelebot/test:latest", shell=True)
        mock_call.assert_any_call("docker push docker.io:8888/skelebot/test:6.6.6", shell=True)
        mock_call.assert_any_call("docker push docker.io:8888/skelebot/test:latest", shell=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_push_omit_version(self, mock_getcwd, mock_call, mock_expanduser):
        host = "docker.io"
        port = 8888
        user = "skelebot"
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        sb.systems.execution.docker.push(config, host, port, user, omit_version=True)

        mock_call.assert_any_call("docker tag test docker.io:8888/skelebot/test:latest", shell=True)
        mock_call.assert_any_call("docker push docker.io:8888/skelebot/test:latest", shell=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_push_omit_latest(self, mock_getcwd, mock_call, mock_expanduser):
        host = "docker.io"
        port = 8888
        user = "skelebot"
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        sb.systems.execution.docker.push(config, host, port, user, omit_latest=True)

        mock_call.assert_any_call("docker tag test docker.io:8888/skelebot/test:6.6.6", shell=True)
        mock_call.assert_any_call("docker push docker.io:8888/skelebot/test:6.6.6", shell=True)

    @mock.patch('skelebot.systems.execution.docker.print')
    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_push_omit_both(self, mock_getcwd, mock_call, mock_expanduser, mock_print):
        host = "docker.io"
        port = 8888
        user = "skelebot"
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        sb.systems.execution.docker.push(config, host, port, user, omit_version=True, omit_latest=True)

        mock_print.assert_called_with("\x1b[33mWARN\x1b[0m | No Tags to Publish")

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_push_host(self, mock_getcwd, mock_call, mock_expanduser):
        host = "docker.io"
        port = 8888
        user = "skelebot"
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        sb.systems.execution.docker.push(config, host, port, user, docker_host='host1')

        mock_call.assert_any_call("docker -H host1 tag test docker.io:8888/skelebot/test:6.6.6", shell=True)
        mock_call.assert_any_call("docker -H host1 tag test docker.io:8888/skelebot/test:latest", shell=True)
        mock_call.assert_any_call("docker -H host1 push docker.io:8888/skelebot/test:6.6.6", shell=True)
        mock_call.assert_any_call("docker -H host1 push docker.io:8888/skelebot/test:latest", shell=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_push_tags(self, mock_getcwd, mock_call, mock_expanduser):
        host = "docker.io"
        port = 8888
        user = "skelebot"
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        sb.systems.execution.docker.push(config, host, port, user, tags=["DEV", "STG"])

        mock_call.assert_any_call("docker tag test docker.io:8888/skelebot/test:6.6.6", shell=True)
        mock_call.assert_any_call("docker push docker.io:8888/skelebot/test:6.6.6", shell=True)
        mock_call.assert_any_call("docker tag test docker.io:8888/skelebot/test:latest", shell=True)
        mock_call.assert_any_call("docker push docker.io:8888/skelebot/test:latest", shell=True)
        mock_call.assert_any_call("docker tag test docker.io:8888/skelebot/test:DEV", shell=True)
        mock_call.assert_any_call("docker push docker.io:8888/skelebot/test:DEV", shell=True)
        mock_call.assert_any_call("docker tag test docker.io:8888/skelebot/test:STG", shell=True)
        mock_call.assert_any_call("docker push docker.io:8888/skelebot/test:STG", shell=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_push_error(self, mock_getcwd, mock_call, mock_expanduser):
        host = "docker.io"
        port = 8888
        user = "skelebot"
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 1

        config = sb.systems.generators.yaml.loadConfig()

        with self.assertRaisesRegex(Exception, "Docker Command Failed"):
            sb.systems.execution.docker.push(config, host, port, user)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.remove')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_build_ephemeral(self, mock_getcwd, mock_call, mock_remove, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        config.ephemeral = True

        sb.systems.execution.docker.build(config, None)
        mock_call.assert_called_once_with("docker build -t test .", shell=True)
        mock_remove.assert_any_call("Dockerfile")
        mock_remove.assert_any_call(".dockerignore")

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_build_non_ephemeral(self, mock_getcwd, mock_call, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()

        sb.systems.execution.docker.build(config, None)
        mock_call.assert_called_once_with("docker build -t test .", shell=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_build_host(self, mock_getcwd, mock_call, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()

        sb.systems.execution.docker.build(config, "root@hosty.McHostface")
        mock_call.assert_called_once_with("docker -H root@hosty.McHostface build -t test .", shell=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_build_with_env(self, mock_getcwd, mock_call, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        config.env = 'test'

        sb.systems.execution.docker.build(config, None)
        mock_call.assert_called_once_with("docker build -t test-test .", shell=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_build_error(self, mock_getcwd, mock_call, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)

        mock_expanduser.return_value = "{path}/test/plugins".format(path=self.path)
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 1

        config = sb.systems.generators.yaml.loadConfig()

        with self.assertRaisesRegex(Exception, "Docker Build Failed"):
            sb.systems.execution.docker.build(config, None)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_run(self, mock_getcwd, mock_call, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        args = Namespace(version='0.1')

        homePath = "{path}/test/plugins".format(path=self.path)
        mock_expanduser.return_value = homePath
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        config.ports = ["1127:1127"]
        job = config.jobs[0]
        job.mappings = ["data/", "/test/output/:/app/output/", "~/temp/:/app/temp/"]
        command = sb.systems.execution.commandBuilder.build(config, job, args)

        expected = "docker run --name test-build --rm -i -p 1127:1127 -v {cwd}/data/:/app/data/ -v /test/output/:/app/output/ -v {path}/temp/:/app/temp/ test /bin/bash -c \"bash build.sh 0.1 --env local --log info\"".format(cwd=folderPath, path=homePath)
        sb.systems.execution.docker.run(config, command, job.mode, config.ports, job.mappings, job.name, host=None)
        mock_call.assert_called_once_with(expected, shell=True)
    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')

    def test_run_host(self, mock_getcwd, mock_call, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        args = Namespace(version='0.1')

        homePath = "{path}/test/plugins".format(path=self.path)
        mock_expanduser.return_value = homePath
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        config.ports = ["1127:1127"]
        job = config.jobs[0]
        job.mappings = ["data/", "/test/output/:/app/output/", "~/temp/:/app/temp/"]
        command = sb.systems.execution.commandBuilder.build(config, job, args)

        expected = "docker -H root@Hosty.McHostface run --name test-build --rm -i -p 1127:1127 -v {cwd}/data/:/app/data/ -v /test/output/:/app/output/ -v {path}/temp/:/app/temp/ test /bin/bash -c \"bash build.sh 0.1 --env local --log info\"".format(cwd=folderPath, path=homePath)
        sb.systems.execution.docker.run(config, command, job.mode, config.ports, job.mappings, job.name, host="root@Hosty.McHostface")
        mock_call.assert_called_once_with(expected, shell=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_run_docker_params(self, mock_getcwd, mock_call, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        memory = 256
        args = Namespace(version='0.1')

        homePath = "{path}/test/plugins".format(path=self.path)
        mock_expanduser.return_value = homePath
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        config.components.append(LimitMemory(memory))
        job = sb.objects.job.Job(name='some_command', help='Dummy', source='echo some_command')
        command = sb.systems.execution.commandBuilder.build(config, job, args)

        expected = "docker run --name test-some_command --rm -i --memory {memory}GB test /bin/bash -c \"echo some_command\"".format(memory=memory)
        sb.systems.execution.docker.run(config, command, job.mode, config.ports, job.mappings, job.name, host=None)
        mock_call.assert_called_once_with(expected, shell=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_run_docker_params_entrypoint(self, mock_getcwd, mock_call, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        memory = 256
        args = Namespace(version='0.1')

        homePath = "{path}/test/plugins".format(path=self.path)
        mock_expanduser.return_value = homePath
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        config.primaryExe = "ENTRYPOINT"
        config.components.append(LimitMemory(memory))
        job = sb.objects.job.Job(name='some_command', help='Dummy', source='echo some_command')
        command = sb.systems.execution.commandBuilder.build(config, job, args)

        expected = "docker run --name test-some_command --rm -i --memory {memory}GB --entrypoint echo test some_command".format(memory=memory)
        sb.systems.execution.docker.run(config, command, job.mode, config.ports, job.mappings, job.name, host=None)
        mock_call.assert_called_once_with(expected, shell=True)

    @mock.patch('os.path.expanduser')
    @mock.patch('skelebot.systems.execution.docker.call')
    @mock.patch('os.getcwd')
    def test_run_gpu(self, mock_getcwd, mock_call, mock_expanduser):
        folderPath = "{path}/test/files".format(path=self.path)
        args = Namespace(version='0.1')

        homePath = "{path}/test/plugins".format(path=self.path)
        mock_expanduser.return_value = homePath
        mock_getcwd.return_value = folderPath
        mock_call.return_value = 0

        config = sb.systems.generators.yaml.loadConfig()
        config.gpu = True
        job = sb.objects.job.Job(name='some_command', help='Dummy', source='echo some_command')
        command = sb.systems.execution.commandBuilder.build(config, job, args)
        
        expected = "docker run --name test-some_command --rm -i --gpus all --ipc=host test /bin/bash -c \"echo some_command\""
        sb.systems.execution.docker.run(config, command, job.mode, config.ports, job.mappings, job.name, host=None)
        mock_call.assert_called_once_with(expected, shell=True)

    @mock.patch('skelebot.systems.execution.docker.call')
    def test_save(self, mock_call):
        mock_call.return_value = 0

        sb.systems.execution.docker.save(sb.objects.config.Config(name="Test Project"), "test.img")
        mock_call.assert_called_once_with("docker save -o test.img test-project", shell=True)

    @mock.patch('skelebot.systems.execution.docker.print')
    @mock.patch('skelebot.systems.execution.docker.call')
    def test_execute_verbose(self, mock_call, mock_print):
        mock_call.return_value = 0

        sb.systems.execution.docker.execute("echo foo", err_msg="Something bad", verbose=True)
        mock_print.assert_called_once_with(Fore.GREEN + "INFO" + Style.RESET_ALL + " | echo foo")
        mock_call.assert_called_once_with("echo foo", shell=True)

if __name__ == '__main__':
    unittest.main()



"""Execution System"""

import sys
from subprocess import call
from ...common import VERSION
from ..scaffolding.scaffolder import Scaffolder
from .commandBuilder import build as buildCommand
from .docker import build as buildDocker
from .docker import run as runDocker
from ...common import INFO

def execute(config, sbParser, args=None):
    """Execute the command(s) that was sent into Skelebot based on the project Config"""

    args = args if args is not None else sys.argv[1:]
    for command in getCommands(args):
        args = sbParser.parseArgs(command)

        if (vars(args).get("version_global", False)):
            print("Skelebot v{}".format(VERSION))
        elif (vars(args).get("contact_global", False)):
            print(config.contact)
        elif (args.job is None):
            sbParser.showHelp()
        elif (args.job == "scaffold"):
            scaffolder = Scaffolder(existing=args.existing)
            scaffolder.scaffold()
        else:
            job = getJob(config, args)

            if (job is not None):
                status = executeJob(config, args, job)
                if status != 0:
                    sys.exit(status)
            else:
                executeComponent(config, args)

def getCommands(args):
    """Split (if needed) and obtain the list of commands that were sent into Skelebot"""

    commands = []
    command = []
    for arg in args:
        if arg == "+":
            commands.append(command)
            command = []
        else:
            command.append(arg)
    commands.append(command)

    return commands

def getJob(config, args):
    """Identify the job in the Config that matches the command that was sent to Skelebot"""

    job = None
    for configJob in config.jobs:
        if args.job == configJob.name:
            job = configJob

    return job

def executeJob(config, args, job):
    """Execute a Config job either natively or through Docker by building a command from it"""

    command = buildCommand(config, job, args, args.native_global)
    host = config.getHost(job=job, args=args)

    is_native = args.native_global
    if (job.native == "always"):
        is_native = True
    elif (job.native == "never"):
        is_native = False

    if (is_native):
        if args.verbose_global:
            print(INFO.format(command))
        status = call(command, shell=True)
    else:
        if (not args.skip_build_global):
            buildDocker(config, host=host, verbose=args.verbose_global)
        ports = sorted(list(set(config.ports + job.ports)))
        status = runDocker(
            config, command, job.mode, ports, job.mappings, job.name, host=host,
            verbose=args.verbose_global
        )
    return(status)

def executeComponent(config, args):
    """Execute a Component of Skelebot"""
    host = config.getHost(args=args)

    for component in config.components:
        if (args.job in component.commands):
            component.execute(config, args, host=host)



import argparse
import copy
import unittest
from unittest import mock

from schema import SchemaError

import skelebot as sb

class TestJupyter(unittest.TestCase):

    jupyter = {
        "port": 123,
        "folder": "test"
    }

    def test_addParsers(self):
        jupyter = sb.components.jupyter.Jupyter(port=1127, folder="notebooks/")

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
        subparsers = parser.add_subparsers(dest="job")
        subparsers = jupyter.addParsers(subparsers)

        self.assertNotEqual(subparsers.choices["jupyter"], None)

    @mock.patch('skelebot.components.jupyter.docker')
    def test_execute(self, mock_docker):
        mock_docker.build.return_value = 0
        config = sb.objects.config.Config()
        args = argparse.Namespace(verbose_global=False)

        jupyter = sb.components.jupyter.Jupyter(port=1127, folder="notebooks/", mappings = ["other_project/data"])
        jupyter.execute(config, args)

        expectedCommand = "jupyter notebook --ip=0.0.0.0 --port=8888 --allow-root --notebook-dir=notebooks/"

        mock_docker.build.assert_called_with(config, host=None, verbose=False)
        mock_docker.run.assert_called_with(
            config, expectedCommand, "it", ["1127:8888"], [".", "other_project/data"],
            "jupyter", host=None, verbose=False
        )

    @mock.patch('skelebot.components.jupyter.docker')
    def test_execute_host(self, mock_docker):
        mock_docker.build.return_value = 0
        config = sb.objects.config.Config()
        args = argparse.Namespace(verbose_global=False)

        jupyter = sb.components.jupyter.Jupyter(port=1127, folder="notebooks/")
        jupyter.execute(config, args, host="host1")

        expectedCommand = "jupyter notebook --ip=0.0.0.0 --port=8888 --allow-root --notebook-dir=notebooks/"

        mock_docker.build.assert_called_with(config, host="host1", verbose=False)
        mock_docker.run.assert_called_with(config, expectedCommand, "it", ["1127:8888"],
            ["."], "jupyter", host="host1", verbose=False
        )

    @mock.patch('skelebot.components.jupyter.docker')
    def test_execute_lab(self, mock_docker):
        mock_docker.build.return_value = 0
        config = sb.objects.config.Config()
        args = argparse.Namespace(verbose_global=False)

        jupyter = sb.components.jupyter.Jupyter(port=1127, folder="notebooks/", mappings=["other_project/data"], lab=True)
        jupyter.execute(config, args)

        expectedCommand = "jupyter lab --ip=0.0.0.0 --port=8888 --allow-root --notebook-dir=notebooks/"

        mock_docker.build.assert_called_with(config, host=None, verbose=False)
        mock_docker.run.assert_called_with(
            config, expectedCommand, "it", ["1127:8888"], [".", "other_project/data"],
            "jupyter", host=None, verbose=False
        )

    def test_validate_valid(self):
        try:
            sb.components.jupyter.Jupyter.validate(self.jupyter)
        except:
            self.fail("Validation Raised Exception Unexpectedly")

    def validate_error(self, attr, reset, expected):
        jupyter = copy.deepcopy(self.jupyter)
        jupyter[attr] = reset

        try:
            sb.components.jupyter.Jupyter.validate(jupyter)
        except SchemaError as error:
            self.assertEqual(error.code, "Jupyter '{attr}' must be a{expected}".format(attr=attr, expected=expected))

    def test_invalid(self):
        self.validate_error('port', "abc", 'n Integer')
        self.validate_error('folder', 123, ' String')

if __name__ == '__main__':
    unittest.main()



"""Command Builder"""

from ...common import EXT_COMMAND

COMMAND = "{pre}{extCommand}{file}{params}{post}"

def build(config, job, args, native=False):
    """Build the command that will either be executed via Docker, or on the native system"""

    ext_list = job.source.split()[0].split(".")
    if len(ext_list) == 1:
        ext = "None"
    else:
        ext = ext_list[-1]

    args = {} if args is None else vars(args)
    arguments = ""
    pre = ""
    post = ""

    # Build the params list from the args and params of the job with the supplied values
    params = buildArgs(job.args, args)
    params += buildParams(job.params, args)
    # Global parameters are ignored, if the direct command is specified in the job source
    if not ext == "None":
        params += buildParams(config.params, args)

    # Construct the pre-run and post-run commands from the components
    for component in config.components:
        preCmd = component.prependCommand(job, native)
        postCmd = component.appendCommand(job, native)
        if (preCmd is not None):
            pre += "{cmd} && ".format(cmd=preCmd)
        if (postCmd is not None):
            post += " && {cmd}".format(cmd=postCmd)

    command = COMMAND.format(extCommand=EXT_COMMAND[ext], file=job.source, args=arguments,
                             params=params, pre=pre, post=post)
    return command

def buildArgs(argParams, args):
    """Build the list of arguments that are passed through to the command"""

    argString = ""
    if (argParams is not None):
        for arg in argParams:
            value = args.get(arg.name.replace("-", "_"))
            if (isinstance(value, str)):
                value = f"'{value}'" if (" " in value) else value

            if (value is not None):
                argString += " {}".format(value)

    return argString

def buildParams(jobParams, args):
    """Build the list of parameters that are passed through to the command"""

    paramString = ""
    if (jobParams is not None):
        for param in jobParams:
            dash_prefix = '--' if len(param.name) > 1 else '-'
            value = args.get(param.name.replace("-", "_"), param.default)
            if (isinstance(value, str)):
                value = f"'{value}'" if (" " in value) else value

            if (param.accepts == "boolean"):
                if (value):
                    paramString += " {dash_prefix}{name}".format(dash_prefix=dash_prefix, name=param.name)
            elif (value is not None):
                if (param.accepts == "list"):
                    value = " ".join(map(str, value))
                paramString += " {dash_prefix}{name} {value}".format(dash_prefix=dash_prefix, name=param.name, value=value)

    return paramString



import argparse
import unittest
from unittest import mock

import skelebot as sb

class TestPlugin(unittest.TestCase):

    def test_addParsers(self):
        plugin = sb.components.plugin.Plugin()

        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
        subparsers = parser.add_subparsers(dest="job")
        subparsers = plugin.addParsers(subparsers)

        self.assertNotEqual(subparsers.choices["plugin"], None)

    @mock.patch('os.path.expanduser')
    @mock.patch('os.path.exists')
    @mock.patch('os.makedirs')
    @mock.patch('skelebot.components.plugin.zipfile.ZipFile')
    def test_execute(self, mock_zipfile, mock_makedirs, mock_exists, mock_expanduser):
        mock_expanduser.return_value = "test/dummy"
        mock_exists.return_value = False

        mock_zip_ref = mock.MagicMock()
        mock_zipfile.return_value.__enter__.return_value = mock_zip_ref
        config = sb.objects.config.Config()
        args = argparse.Namespace(plugin="test.zip")

        plugin = sb.components.plugin.Plugin()
        plugin.execute(config, args)

        mock_zipfile.assert_called_once_with("test.zip", "r")
        mock_zip_ref.extractall.assert_called_once_with("test/dummy")

        mock_makedirs.assert_any_call("test/dummy", exist_ok=True)

if __name__ == '__main__':
    unittest.main()



import argparse
import unittest
from unittest import mock

import skelebot as sb

class TestDexec(unittest.TestCase):

    def test_addParsers(self):
        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
        subparsers = parser.add_subparsers(dest="job")
        dexec = sb.components.dexec.Dexec()
        subparsers = dexec.addParsers(subparsers)

        self.assertNotEqual(subparsers.choices["exec"], None)

    @mock.patch('skelebot.components.dexec.docker')
    def test_execute_nomap(self, mock_docker):
        config = sb.objects.config.Config(name="test-dexec")
        args = argparse.Namespace(map=False, verbose_global=False)

        dexec = sb.components.dexec.Dexec()
        dexec.execute(config, args)

        mock_docker.build.assert_called_with(config, host=None, verbose=False)
        mock_docker.run.assert_called_with(
            config, "/bin/bash", "it", [], [], "exec", host=None, verbose=False
        )

    @mock.patch('skelebot.components.dexec.docker')
    def test_execute_map(self, mock_docker):
        config = sb.objects.config.Config(name="test-dexec")
        args = argparse.Namespace(map=True, verbose_global=False)

        dexec = sb.components.dexec.Dexec()
        dexec.execute(config, args)

        mock_docker.build.assert_called_with(config, host=None, verbose=False)
        mock_docker.run.assert_called_with(
            config, "/bin/bash", "it", [], ["."], "exec", host=None, verbose=False
        )

if __name__ == '__main__':
    unittest.main()
